//	2  ذ ִ Ǻ : ax^2 + bx + c=0
//	 װ  Ǽ ذ °?
//	x ^ 2 + 5x + 9 = 0
//	2x ^ 2 + 6x + 0 = 0
//	3x ^ 2 + 7x + 1 = 0
//	4x ^ 2 + 8x + 2 = 0

#include <cstdio>
#include <iostream>
#include <ppl.h>
#include <amp.h>
using namespace std;
using namespace concurrency;

int a[4] = { 1, 2, 3, 4 };
int b[4] = { 5, 6, 7, 8 };
int c[4] = { 9, 0, 1, 2 };

//	C code : keyword - "for"
void step1(int r[4])
{
	for (int i = 0; i < 4; i++)
	{
		r[i] = b[i] * b[i] - 4 * a[i] * c[i];
	}
}

//	GPU code : keyword - "array_view", "synchronize"
void step2(int r[4])
{
	concurrency::array_view<int, 1>  ga(4, a);
	concurrency::array_view<int, 1>  gb(4, b);
	concurrency::array_view<int, 1>  gc(4, c);
	concurrency::array_view<int, 1>  gr(4, r);

	concurrency::parallel_for_each(gr.extent, [=](index<1> i) restrict(amp)
	{
		gr[i] = gb[i] * gb[i] - 4 * ga[i] * gc[i];
	});
	gr.synchronize();
}

//	GPU code : keyword - "const", "discard_data"
void step3(int r[4])
{
	concurrency::array_view<const int, 1>  ga(4, a);
	concurrency::array_view<const int, 1>  gb(4, b);
	concurrency::array_view<const int, 1>  gc(4, c);
	concurrency::array_view<int, 1>  gr(4, r);
	gr.discard_data();

	concurrency::parallel_for_each(gr.extent, [=](index<1> i) restrict(amp)
	{
		gr[i] = gb[i] * gb[i] - 4 * ga[i] * gc[i];
	});
	gr.synchronize();
}

//	GPU code : keyword - "tile", "tiled_index", "tile_static"
void step4(int r[4])
{
	concurrency::array_view<const int, 1>  ga(4, a);
	concurrency::array_view<const int, 1>  gb(4, b);
	concurrency::array_view<const int, 1>  gc(4, c);
	concurrency::array_view<int, 1>  gr(4, r);
	gr.discard_data();

	concurrency::parallel_for_each(gr.extent.tile<1>(), [=](concurrency::tiled_index<1> i) restrict(amp)
	{
		tile_static int local_b;
		local_b = gb[i.global];
		gr[i] = local_b * local_b - 4 * ga[i] * gc[i];
	});
	gr.synchronize();
}

int main()
{
	int r[4];
	auto print_r = [&r](std::string s)
	{
		cout << s << endl;
		for (int i = 0; i < 4; i++)
		{
			if (r[i] < 0)	cout << i << "th result is incorrent : " << r[i] << endl;
			else cout << i << "th result is OK : " << r[i] << endl;
		}
	};

	step1(r);	print_r("step1 result");
	step2(r);	print_r("step2 result");
	step3(r);	print_r("step3 result");
	step4(r);	print_r("step4 result");
}
